import asyncio
from typing import Awaitable, Callable, TypeVar
from . import protocol
from datetime import datetime
from bleak import BleakClient
from bleak.backends.device import BLEDevice
from bleak_retry_connector import (
    establish_connection,
    BleakClientWithServiceCache
)

class UnexpectedResponseError(Exception):
    pass

T = TypeVar('T')
EncoderOrMessage = Callable[[BleakClient], Awaitable[bytearray]] | bytearray
Decoder = Callable[[bytearray], T]

def decode_response_success(response: bytearray) -> bool:
    return protocol.decode_response_status(response).is_success()

class ZuliSmartplug():

    def __init__(self,
        address: str,
        device: BLEDevice | Callable[[str], Awaitable[BLEDevice]],
        num_retries: int = 3
    ):
        self._address = address
        self._device: BLEDevice | Callable[[str], Awaitable[BLEDevice]] = device
        self._client: BleakClient | None = None
        self._lock = asyncio.Lock()
        self._num_retries = num_retries

    async def __get_device(self) -> BLEDevice:
        if callable(self._device):
            return await self._device(self._address)
        else:
            return self._device

    async def __get_connected_client(self) -> BleakClient:
        if self._client and self._client.is_connected:
            return self._client
        else:
            device_handle = await self.__get_device()
            return await establish_connection(
                BleakClientWithServiceCache,
                device_handle,
                self._address,
                max_attempts=self._num_retries
            )
    
    async def _send_command(
        self,
        encode_message: EncoderOrMessage,
        decode_response: Decoder[T] = decode_response_success
    ) -> T:
        async with self._lock:
            client = await self.__get_connected_client()

            # Encode and send message
            message = await encode_message(client) if callable(encode_message) else encode_message
            await client.write_gatt_char(protocol.COMMAND_PIPE, data=message,
                                        response=True)
            
            # Read and decode response
            raw_response = await client.read_gatt_char(protocol.COMMAND_PIPE)
            try:
                return decode_response(raw_response)
            except Exception as e:
                raise UnexpectedResponseError() from e
    
    async def disconnect(self):
        if self._client != None and self._client.is_connected:
            await self._client.disconnect()

    @property
    def address(self):
        return self._address
    
    async def on(self, brightness: int):
        return await self._send_command(protocol.encode_on(brightness))

    async def off(self):
        return await self._send_command(protocol.encode_off())
    
    async def read(self):
        return await self._send_command(
            protocol.encode_read(),
            protocol.decode_read
        )
    
    async def get_mode(self):
        return await self._send_command(
            protocol.encode_get_mode(),
            protocol.decode_get_mode
        )

    async def set_mode(self, is_appliance: bool):
        return await self._send_command(protocol.encode_set_mode(is_appliance))

    async def sync_clock(self):
        return await self._send_command(protocol.encode_set_clock(datetime.today()))
    
    async def get_clock(self):
        return await self._send_command(
            protocol.encode_get_clock(),
            protocol.decode_get_clock
        )

    async def read_power(self):
        return await self._send_command(
            protocol.encode_read_power(),
            protocol.decode_read_power
        )
    
    async def add_schedule(self, schedule: protocol.Schedule):
        return await self._send_command(protocol.encode_add_schedule(schedule))

    async def get_schedule(self, i: int):
        return await self._send_command(
            protocol.encode_get_schedule(i),
            protocol.decode_get_schedule
        )

    async def get_schedule_info(self):
        return await self._send_command(
            protocol.encode_get_schedule_info(),
            protocol.decode_get_schedule_info
        )

    async def get_schedules_stream(self):
        schedule_info = await self.get_schedule_info()
        num_schedules = schedule_info['num_schedules']

        for i in range(1, num_schedules + 1):
            yield await self.get_schedule(i)
    
    async def get_schedules(self):
        schedules = [schedule async for schedule in self.get_schedules_stream()]
        return sorted(schedules, key=lambda s: s.id)

    async def remove_schedule(self, i: int):
        schedule = await self.get_schedule(i)
        return await self._send_command(protocol.encode_remove_schedule(schedule))



"""Zuli protocol implementation"""
import datetime
from enum import Enum
from typing import TypedDict

ZULI_SERVICE = '04ee929b-bb13-4e77-8160-18552daf06e1'
COMMAND_PIPE = 'ffffff03-bb13-4e77-8160-18552daf06e1'

class ZuliCommand(Enum):
    RESET = 2
    VERSION_READ = 6
    FLAGS_READ = 7
    CLOCK_SET = 8
    CLOCK_GET = 9
    NETWORK_SET = 10
    NETWORK_GET = 11
    MODE_SET = 16
    MODE_GET = 17
    ATTRIBUTE_SET = 21
    ATTRIBUTE_GET = 22
    ON = 23
    OFF = 24
    READ = 25
    POWER_READ = 32
    ENERGY_READ_INFO = 33
    ENERGY_READ_ACCUM = 34
    ENERGY_READ_LATCH = 35
    ENERGY_LATCH_RESET_ALL = 36
    SCHEDULE_INFO_GET = 48
    SCHEDULE_GET = 49
    SCHEDULE_ENABLE = 50
    SCHEDULE_ADD = 51
    SCHEDULE_REMOVE = 52
    SCHEDULE_REMOVE_ALL = 53
    DEFAULT_OUTPUT_SET = 80
    DEFAULT_OUTPUT_GET = 81
    BOOKMARK = 126
    BATCH = 127

class ZuliStatus(Enum):
    SUCCESS = 0
    BUSY = 5
    INVALID_PARAM = 6
    BAD_LENGTH = 15
    ALREADY_SET = 9

    def is_success(self) -> bool:
        return self == ZuliStatus.SUCCESS or self == ZuliStatus.ALREADY_SET

def decode_response_status(response: bytearray) -> ZuliStatus:
    """Returns response status"""
    return ZuliStatus(response[1])

def encode_read() -> bytearray:
    """Message to query the state of a smartplug (on/off/brightness)"""
    return bytearray([ZuliCommand.READ.value])

def decode_read(response: bytearray) -> int:
    """Returns the current brightness of the device as an integer from 0-100 
    (for appliances, this value will be either 0 for off or 100 for on)"""
    return int(response[2])

def encode_on(brightness: int = 0) -> bytearray:
    """Message to turn a smartplug on, optionally at a specified brightness
    
    :param brightness: a number between 0 and 100 (otherwise will be trimmed);
        note that this defaults to 0, which is functionally equivalent to 100;
        brightness is ignored by the smartplug when in appliance mode"""
    brightness = min(100, max(0, brightness))
    return bytearray([ZuliCommand.ON.value, 0, 0, 0, 0, brightness, 0, 0, 0])

def encode_off() -> bytearray:
    """Message to turn a smartplug off"""
    return bytearray([ZuliCommand.OFF.value, 0, 0, 0])

def encode_get_mode() -> bytearray:
    """Message to query the mode of a smartplug (appliance or dimmable)"""
    return bytearray([ZuliCommand.MODE_GET.value])

def decode_get_mode(response: bytearray) -> bool:
    """Returns whether the device is in appliance mode (True=appliance,
    False=dimmable)"""
    return not bool(response[2])

def encode_set_mode(is_appliance: bool = True) -> bytearray:
    """Message to set the mode of a smartplug
    
    :param is_appliance: by default True, indicating that the smartplug is
        attached to a high power device that does not support dimming (good for
        appliances, non-dimmable lights, etc.); otherwise, the smartplug allows
        dimming
    """
    mode = 0 if is_appliance else 1
    return bytearray([ZuliCommand.MODE_SET.value, mode])

def encode_set_clock(time: datetime.datetime) -> bytearray:
    """Message to set the clock of a smartplug

    Smartplugs maintain system time for schedules, though the time does not
    persist past power cycles.
    """
    year = time.year.to_bytes(2)
    weekday = ((time.weekday() + 1) % 7) + 1
    return bytearray([ZuliCommand.CLOCK_SET.value, year[0], year[1], time.month, time.day,
                      weekday, time.hour, time.minute, time.second])

def encode_get_clock() -> bytearray:
    """Message to to poll the current system time on a smartplug"""
    return bytearray([ZuliCommand.CLOCK_GET.value])

def decode_get_clock(response: bytearray) -> datetime.datetime:
    """Produces a datetime object from a get clock packet and fails if the
    packet is malformed"""
    year = int.from_bytes(response[2:4])
    return datetime.datetime(year, month=response[4], day=response[5],
                             hour=response[7], minute=response[8],
                             second=response[9])

def encode_read_power() -> bytearray:
    """Message to to read current power consumption"""
    return bytearray([ZuliCommand.POWER_READ.value])

class Power(TypedDict):
    irms_ma: int
    power_mw: int
    power_factor: int
    voltage_mv: int

def decode_read_power(response: bytearray) -> Power:
    """Returns the current power consumption in watts from a read power packet
    and fails if the packet is malformed"""
    irms_ma = int.from_bytes(response[2:4])
    power_mw = int.from_bytes(response[4:7])
    power_factor = int.from_bytes(response[7:9])
    voltage_mv = int.from_bytes(response[9:12])
    return Power(irms_ma=irms_ma, power_mw=power_mw, power_factor=power_factor,
                 voltage_mv=voltage_mv)

class ScheduleAction(Enum):
    ON = 1
    OFF = 2

class Schedule():
    """A representation of a schedule that can be used to turn a smartplug on
    or off at a specific time"""
    WEEKDAY_SYMBOL = "MTWTFSS"

    def __init__(self, time: datetime.time, id: int = 0,
                 action: ScheduleAction = ScheduleAction.ON,
                 weekdays: list[bool] = [True] * 7, enabled: bool = True,
                 schedule_id: int = 0):
        self.id = id
        self.action = action
        self.time = time
        self.weekdays = weekdays
        self.enabled = enabled
        self.schedule_id = schedule_id

    @staticmethod
    def from_bytes(raw: bytearray):
        id = raw[0]
        action = ScheduleAction(raw[1])
        time = datetime.time(hour=raw[4], minute=raw[5], second=raw[6])
        # Represented from index 0 == Monday ... 6 == Sunday as does datetime,
        # note that this is a departure from the way the smartplugs consider
        # Sunday to be the first day of the week
        weekdays = []
        for i in range(7):
            # Account the start of the week difference between the protocol
            # and Python datetime
            i = (i + 1) % 7
            bitflag = 1 << i
            weekdays.append(raw[7] & bitflag == bitflag)
        enabled = raw[8] == 1
        schedule_id = raw[9]
        return Schedule(time, id=id, action=action, weekdays=weekdays,
                        enabled=enabled, schedule_id=schedule_id)

    def to_bytes(self) -> bytearray:
        weekdays = 0
        for i in range(7):
            # Account the start of the week difference between the protocol
            # and Python datetime
            shift = (i + 1) % 7
            bitflag = self.weekdays[i] << shift
            weekdays |= bitflag
        return bytearray([self.id, self.action.value, 0, 0, self.time.hour,
                      self.time.minute, self.time.second, weekdays,
                      self.enabled, self.schedule_id])
    
    def without_id(self) -> bytearray:
        """Returns a trimmed byte representation of the schedule without
        identifiers, useful when removing schedules"""
        raw = self.to_bytes()
        return raw[1:8]
    
    def __str__(self):
        weekdays_sym = map(lambda i : self.WEEKDAY_SYMBOL[i]
                           if self.weekdays[i] else "-", range(7))
        weekdays_str = " ".join(weekdays_sym)
        action_str = "Turn On" if self.action == ScheduleAction.ON else "Turn Off"
        enabled_str = "Enabled" if self.enabled else "Disabled"
        return f"{action_str}  {weekdays_str}  at {self.time.isoformat()} ({enabled_str})"

def encode_add_schedule(schedule: Schedule) -> bytearray:
    """Message to to push a new schedule to the smartplug
    """
    packet = bytearray([ZuliCommand.SCHEDULE_ADD.value])
    packet.extend(schedule.to_bytes())
    return packet

def encode_get_schedule(i: int) -> bytearray:
    """Message to to get a single schedule saved to the smartplug

    This packet will typically be sent n times, n being the number of schedules
    saved to the smartplug, after first sending a get schedule info packet to
    determine the value of n.
    
    :param i: a number between 1 and the number of schedules on the smartplug;
        the number that corresponds to a specific schedule does not stay the
        same between operations that change schedules
    """
    return bytearray([ZuliCommand.SCHEDULE_GET.value, i])

def decode_get_schedule(response: bytearray) -> Schedule:
    """Returns a single schedule from a get schedule packet and fails if the
    packet is malformed"""
    return Schedule.from_bytes(response[2:])

def encode_get_schedule_info() -> bytearray:
    """Message to to get schedule info"""
    return bytearray([ZuliCommand.SCHEDULE_INFO_GET.value, 0])

class ScheduleInfo(TypedDict):
    num_schedules: int
    max_schedules: int

def decode_get_schedule_info(response: bytearray) -> ScheduleInfo:
    """Returns a tuple of the number of events and the maximum supported number
    and fails if the packet is malformed"""
    return ScheduleInfo(num_schedules=response[2], max_schedules=response[3])

def encode_remove_schedule(schedule: Schedule) -> bytearray:
    """Message to to remove a single schedule saved to the smartplug"""
    packet = bytearray([ZuliCommand.SCHEDULE_REMOVE.value, 0])
    packet.extend(schedule.without_id())
    return packet

def encode_remove_all_schedules() -> bytearray:
    """Untested. Reconstructed from Zuli Android app"""
    packet = bytearray([ZuliCommand.SCHEDULE_REMOVE_ALL.value])
    confirm_remove_all = 46140
    packet.extend(confirm_remove_all.to_bytes(2))
    return packet

def encode_read_energy_info() -> bytearray:
    """Untested. Reconstructed from Zuli Android app"""
    return bytearray([ZuliCommand.ENERGY_READ_INFO.value, 0])

class EnergyInfo(TypedDict):
    a: int
    b: int
    c: int
    d: int

def decode_read_energy_info(response: bytearray) -> EnergyInfo:
    a = response[2]
    b = response[4]
    c = int.from_bytes(response[5:7])
    d = int.from_bytes(response[7:9])
    return EnergyInfo(a=a, b=b, c=c, d=d)

def encode_latch_data(latch_id: int) -> bytearray:
    """Untested. Reconstructed from Zuli Android app"""
    packet = bytearray([ZuliCommand.ENERGY_READ_LATCH.value, 0])
    packet.extend(latch_id.to_bytes(2))
    return packet

class LatchData(TypedDict):
    value: int
    duration: int
    unix_time_sec: int
    unix_time_ms: int

def decode_read_latch_data(response: bytearray) -> LatchData:
    """Untested. Reconstructed from Zuli Android app"""
    value = int.from_bytes(response[2:9])
    duration = int.from_bytes(response[9:14])
    unix_time_sec = int.from_bytes(response[14:18])
    unix_time_ms = int.from_bytes(response[18:20])
    return LatchData(value=value, duration=duration,
                     unix_time_sec=unix_time_sec, unix_time_ms=unix_time_ms)

def encode_reset_all_latches(num_latches: int) -> bytearray:
    """Untested. Reconstructed from Zuli Android app"""
    packet = bytearray([ZuliCommand.ENERGY_LATCH_RESET_ALL.value, 0])
    packet.extend(num_latches.to_bytes(2))
    confirm_reset_all = 5693
    packet.extend(confirm_reset_all.to_bytes(2))
    return packet

def encode_reset_plug() -> bytearray:
    """Untested. Reconstructed from Zuli Android app"""
    packet = bytearray([ZuliCommand.RESET.value, ZuliCommand.RESET.value, 0])
    confirm_reset = 22890
    packet.extend(confirm_reset.to_bytes(2))
    return packet



from typing import Awaitable, Callable, TypeVar
from . import protocol
from .smartplug import ZuliSmartplug
import asyncio
import sys
import argparse
from datetime import time
from bleak import BleakScanner

class InteractiveArgumentParser(argparse.ArgumentParser):
    def exit(self, status=0, message=None): # pyright: ignore[reportIncompatibleMethodOverride]
        if message:
            raise argparse.ArgumentError(argument=None, message=message)

def filter_devices(devices: dict[str, ZuliSmartplug], addresses: list[str]):
    # An empty list of addresses returns all devices
    if len(addresses) == 0:
        return list(devices.values())
    
    # Find devices in dict with (possible) partial addresses
    # Performance is not great, but the number of devices should always be very
    # low
    filtered = []
    for addr in addresses:
        for k, v in devices.items():
            if k.startswith(addr):
                filtered.append(v)
    return filtered

T = TypeVar('T')
    
def wrap_method(smartplug_func: Callable[[ZuliSmartplug], Callable[..., Awaitable[T]]]):
    """Creates and returns a wrapper method that reads command line arguments
    and then calls a specified method in the smartplug module."""
    async def do(args: argparse.Namespace, devices: dict[str, ZuliSmartplug]):
        command_devices = filter_devices(devices, args.devices)
        method_params = args.params(args) if hasattr(args, 'params') else []

        async def command_for_device(
            device: ZuliSmartplug,
            smartplug_func: Callable[..., Awaitable[T]],
            args: list
        ):
            return (device, await smartplug_func(*args))

        pending_commands = [command_for_device(device, smartplug_func(device), method_params) for device in command_devices]
        for result_future in asyncio.as_completed(pending_commands):
            result = await result_future
            device = result[0] if isinstance(result, tuple) else None
            result = result[1] if isinstance(result, tuple) else result

            if isinstance(result, list):
                results = result
                result = ""
                for item in results:
                    result += f"{item}\n"
                result = result.strip()
            
            if device != None:
                if isinstance(result, str) and '\n' in result:
                    print(f"\n==== {device.address} ====\n")
                    print(result)
                else:
                    print(f"{device.address} : {result}")
            else:
                print(result)            
    return do
    
async def list_devices(args: argparse.Namespace,
                    devices: dict[str, ZuliSmartplug]):
    for client in devices.values():
        print(client.address)

async def ainput(prompt: str) -> str:
    print(f"{prompt} ", end='', flush=True)
    return (await asyncio.to_thread(sys.stdin.readline)).rstrip('\n')
    
def configure_parser():
    parser = InteractiveArgumentParser(prog="zuli", exit_on_error=False)
    subparsers = parser.add_subparsers()

    parent_parser = InteractiveArgumentParser(add_help=False)
    parent_parser.add_argument('-d', '--devices', action='extend', nargs='+',
                            type=str, default=[])

    parser_on = subparsers.add_parser('on', parents=[parent_parser])
    parser_on.add_argument('brightness', nargs='?', default=0, type=int)
    parser_on.set_defaults(func=wrap_method(lambda a: a.on),
                            params=lambda a : [a.brightness])

    parser_off = subparsers.add_parser('off', parents=[parent_parser])
    parser_off.set_defaults(func=wrap_method(lambda a: a.off))

    parser_off = subparsers.add_parser('read', parents=[parent_parser])
    parser_off.set_defaults(func=wrap_method(lambda a: a.read))

    parser_off = subparsers.add_parser('mode', parents=[parent_parser])
    parser_off.set_defaults(func=wrap_method(lambda a: a.get_mode))

    parser_mode = subparsers.add_parser('set_mode', parents=[parent_parser])
    parser_mode.add_argument('mode', choices=['dimmable', 'appliance'])
    parser_mode.set_defaults(func=wrap_method(lambda a: a.set_mode),
                                params=lambda a : [a.mode == 'appliance'])
    
    parser_power = subparsers.add_parser('power', parents=[parent_parser])
    parser_power.set_defaults(func=wrap_method(lambda a: a.read_power))

    parser_time = subparsers.add_parser('time', parents=[parent_parser])
    parser_time.set_defaults(func=wrap_method(lambda a: a.get_clock))

    parser_synctime = subparsers.add_parser('synctime',
                                            parents=[parent_parser])
    parser_synctime.set_defaults(func=wrap_method(lambda a: a.sync_clock))

    parser_schedule = subparsers.add_parser('schedules',
                                            parents=[parent_parser])
    parser_schedule.set_defaults(func=wrap_method(lambda a: a.get_schedules))

    parser_schedule_remove = subparsers.add_parser('remove_schedule',
                                                parents=[parent_parser])
    parser_schedule_remove.add_argument('schedule', type=int)
    parser_schedule_remove.set_defaults(func=wrap_method(lambda a: a.remove_schedule),
                                        params=lambda a : [a.schedule])
    
    parser_schedule_add = subparsers.add_parser('add_schedule',
                                                parents=[parent_parser])
    parser_schedule_add.add_argument('time', type=time.fromisoformat)
    parser_schedule_add.add_argument('action', choices=['on', 'off'])
    def schedule_params(a):
        return [protocol.Schedule(time=a.time, action=protocol.ScheduleAction.OFF
                                            if a.action == "off"
                                            else protocol.ScheduleAction.ON)]
    parser_schedule_add.set_defaults(func=wrap_method(lambda a: a.add_schedule),
                                        params=schedule_params)
    
    parser_devices = subparsers.add_parser('devices')
    parser_devices.set_defaults(func=list_devices)

    subparsers.add_parser('quit')

    return parser

async def main():
    devices: dict[str, ZuliSmartplug] = {}

    async def discover():
        try:
            async with BleakScanner(service_uuids=[protocol.ZULI_SERVICE]) as scanner:
                async for (device, advertisement_data) in scanner.advertisement_data():
                    if device.address not in devices:
                        zuli_device = ZuliSmartplug(device.address, device)
                        devices[device.address] = zuli_device
        except asyncio.CancelledError:
            return
    discovery_task = asyncio.create_task(discover())

    try:
        print("Ready. Discovery will continue in the background")
        parser = configure_parser()

        while True:
            command = await ainput(">>>")
            if command.startswith("quit"):
                break

            try:
                args = parser.parse_args(command.split())
                if hasattr(args, 'func'):
                    await args.func(args, devices)
            except Exception as e:
                print(e)
    except Exception as e:
        print(e)
    finally:
        print(f"Stopping discovery and closing connections")
        discovery_task.cancel()
        await discovery_task
        await asyncio.gather(*[device.disconnect()
                                for device in devices.values()])

asyncio.run(main())
